@@ -39,14 +39,17 @@ programmatically.  A document can have multiple indexes by adding extra
 
 '''
 import pymongo
+from bson import DBRef
 from pymongo import GEO2D
 from collections import defaultdict
 from mongoalchemy.util import classproperty, UNSET
 from mongoalchemy.query_expression import QueryField
-from mongoalchemy.fields import ObjectIdField, Field, BadValueException, SCALAR_MODIFIERS
+from mongoalchemy.fields import (ObjectIdField, Field, BadValueException, 
+                                 SCALAR_MODIFIERS, DocumentField)
 from mongoalchemy.exceptions import DocumentException, MissingValueException, ExtraValueException, FieldNotRetrieved, BadFieldSpecification
 
 document_type_registry = defaultdict(dict)
+collection_registry = defaultdict(dict)
 
 class DocumentMeta(type):
     def __new__(mcs, classname, bases, class_dict):
@@ -87,29 +90,33 @@ class DocumentMeta(type):
                 continue
             new_class._fields[name] = maybefield
         
-        # 3. register type
-        if new_class.config_namespace != None:
-            name = new_class.config_full_name
-            if name == None:
-                name = new_class.__name__
-            document_type_registry[new_class.config_namespace][name] = new_class
-
-        # 5. Add proxies
-        for name, field in new_class.get_fields().iteritems():
-            if field.proxy is not None:
-                setattr(new_class, field.proxy, Proxy(name))
-            if field.iproxy is not None:
-                setattr(new_class, field.iproxy, IProxy(name, field.ignore_missing))
-        
-        # 4.  Add subclasses
-
+        # 3.  Add subclasses
         for b in bases:
             if 'Document' in globals() and issubclass(b, Document):
                 b.add_subclass(new_class)
             if not hasattr(b, 'config_polymorphic_collection'):
                 continue
-            if b.config_polymorphic_collection and 'config_collection' not in class_dict:
+            if b.config_polymorphic_collection and 'config_collection_name' not in class_dict:
                 new_class.config_collection_name = b.get_collection_name()
+        
+        # 4. register type
+        if new_class.config_namespace is not None:
+            name = new_class.config_full_name
+            if name == None:
+                name = new_class.__name__
+
+            ns = new_class.config_namespace
+            document_type_registry[ns][name] = new_class
+
+            # if the new class uses a polymorphic collection we should only
+            # set up the collection name to refer to the base class
+            # TODO: if non-polymorphic classes use the collection registry they
+            # will just overwrite for now.
+            collection = new_class.get_collection_name()
+            current = collection_registry[ns].get(collection)
+            if current is None or issubclass(current, new_class):
+                collection_registry[ns][collection] = new_class
+
 
 
         return new_class
@@ -179,25 +186,27 @@ class Document(object):
         self.partial = retrieved_fields is not None
         self.retrieved_fields = self.__normalize(retrieved_fields)
         
-        self._dirty = {}
-        
-        self._field_values = {}
+        # Mapping from attribute names to values.
+        self._values = {}
         self.__extra_fields = {}
         
         cls = self.__class__
                 
         fields = self.get_fields()
         for name, field in fields.iteritems():
+            # print name
             if self.partial and field.db_field not in self.retrieved_fields:
-                continue
-            
-            if name in kwargs:
-                getattr(cls, name).set_value(self, kwargs[name], from_db=loading_from_db)
-                continue
-            # DO I NEED THIS?
-            # elif field.default != UNSET:
-            #     getattr(cls, name).set_value(self, field.default, from_db=loading_from_db)
-
+                self._values[name] = Value(field, self, retrieved=False)
+            elif name in kwargs:
+                field = getattr(cls, name)
+                value = kwargs[name]
+                self._values[name] = Value(field, self, 
+                                           from_db=loading_from_db)
+                getattr(cls, name).set_value(self, kwargs[name])
+            elif field.auto:
+                self._values[name] = Value(field, self, from_db=False)
+            else:
+                self._values[name] = Value(field, self, from_db=False)
         
         for k in kwargs:
             if k not in fields:
@@ -253,6 +262,8 @@ class Document(object):
             return self.mongo_id == other.mongo_id
         except:
             return False
+    def __ne__(self, other):
+        return not self.__eq__(other)
 
     def get_dirty_ops(self, with_required=False):
         ''' Returns a dict with the update operations necessary to make the 
@@ -270,7 +281,7 @@ class Document(object):
                 continue
             dirty_ops = field.dirty_ops(self)
             if not dirty_ops and with_required and field.required:
-                dirty_ops = field.update_ops(self)
+                dirty_ops = field.update_ops(self, force=True)
                 if not dirty_ops:
                     raise MissingValueException(name)
             
@@ -362,7 +373,10 @@ class Document(object):
         except AttributeError:
             return False
         return True 
-    
+    def to_ref(self, db=None):
+        return DBRef(id=self.mongo_id, 
+                     collection=self.get_collection_name(),
+                     database=db)
     def wrap(self):
         ''' Returns a transformation of this document into a form suitable to 
             be saved into a mongo database.  This is done by using the ``wrap()``
@@ -382,20 +396,7 @@ class Document(object):
                 if field.required:
                     raise MissingValueException(name)
         return res
-    
-    @classmethod
-    def validate_unwrap(cls, obj, fields=None, session=None):
-        ''' Attempts to unwrap the document, and raises a BadValueException if
-            the operation fails. A TODO is to make this function do the checks
-            without actually doing the (potentially expensive) 
-            deserialization'''
-        try:
-            cls.unwrap(obj, fields=fields, session=session)
-        except Exception, e:
-            e_type = type(e).__name__
-            cls_name = cls.__name__
-            raise BadValueException(cls_name, obj, '%s Exception validating document' % e_type, cause=e)
-    
+        
     @classmethod
     def unwrap(cls, obj, fields=None, session=None):
         ''' Returns an instance of this document class based on the mongo object 
@@ -442,7 +443,7 @@ class Document(object):
         if fields is not None:
             params['retrieved_fields'] = fields
         obj = cls(loading_from_db=True, **params)
-        obj.__mark_clean()
+        obj._mark_clean()
         obj._session = session
         return obj
 
@@ -452,8 +453,10 @@ class Document(object):
     def _set_session(self, session):
         self._session = session
 
-    def __mark_clean(self):
-        self._dirty.clear()
+    def _mark_clean(self):
+        # print 'CLEAR DIRTY'
+        for k, v in self._values.iteritems():
+            v.clear_dirty()
         
 
 class DictDoc(object):
@@ -463,8 +466,8 @@ class DictDoc(object):
     '''
     def __getitem__(self, name):
         ''' Gets the field ``name`` from the document '''
-        fields = self.get_fields()
-        if name in fields:
+        # fields = self.get_fields()
+        if name in self._values:
             return getattr(self, name)
         raise KeyError(name)
     
@@ -496,100 +499,6 @@ class DictDoc(object):
         return True
         
 
-class DocumentField(Field):
-    ''' A field which wraps a :class:`Document`'''
-    
-    has_subfields = True
-    has_autoload = True
-    
-    def __init__(self, document_class, **kwargs):
-        super(DocumentField, self).__init__(**kwargs)
-        self.__type = document_class
-    
-    @property
-    def type(self):
-        if not isinstance(self.__type, basestring) and issubclass(self.__type, Document):
-            return self.__type
-        if self.parent.config_namespace == None:
-            raise BadFieldSpecification('Document namespace is None.  Strings are not allowed for DocumentFields')
-        type = document_type_registry[self.parent.config_namespace].get(self.__type)
-        if type == None or not issubclass(type, Document):
-            raise BadFieldSpecification('No type found for %s.  Maybe it has not been imported yet and is not registered?' % self.__type)
-        return type
-    
-    def dirty_ops(self, instance):
-        ''' Returns a dict of the operations needed to update this object.  
-            See :func:`Document.get_dirty_ops` for more details.'''
-        try:
-            document = getattr(instance, self._name)
-        except AttributeError:
-            return {}
-        if len(document._dirty) == 0 and \
-           self.__type.config_extra_fields != 'ignore':
-            return {}
-        
-        ops = document.get_dirty_ops()
-        
-        ret = {}
-        for op, values in ops.iteritems():
-            ret[op] = {}
-            for key, value in values.iteritems():
-                name = '%s.%s' % (self._name, key)
-                ret[op][name] = value
-        return ret
-    
-    def subfields(self):
-        ''' Returns the fields that can be retrieved from the enclosed 
-            document.  This function is mainly used internally'''
-        return self.type.get_fields()
-    
-    def sub_type(self):
-        return self.type
-    
-    def is_valid_unwrap(self, value, fields=None):
-        ''' Called before wrapping.  Calls :func:`~DocumentField.is_valid_unwrap` and 
-            raises a :class:`BadValueException` if validation fails            
-            
-            :param value: The value to validate
-            :param fields: The fields being returned if this is a partial \
-                document. They will be ignored when validating the fields \
-                of ``value``
-        '''
-        try:
-            self.validate_unwrap(value, fields=fields)
-        except BadValueException, bve:
-            return False
-        return True
-    
-    def wrap(self, value):
-        ''' Validate ``value`` and then use the document's class to wrap the 
-            value'''
-        self.validate_wrap(value)
-        return self.type.wrap(value)
-    
-    def unwrap(self, value, fields=None, session=None):
-        ''' Validate ``value`` and then use the document's class to unwrap the 
-            value'''
-        self.validate_unwrap(value, fields=fields, session=session)
-        return self.type.unwrap(value, fields=fields, session=session)
-    
-    def validate_wrap(self, value):
-        ''' Checks that ``value`` is an instance of ``DocumentField.type``.
-            if it is, then validation on its fields has already been done and
-            no further validation is needed.
-        '''
-        if not isinstance(value, self.type):
-            self._fail_validation_type(value, self.type)
-    
-    def validate_unwrap(self, value, fields=None, session=None):
-        ''' Validates every field in the underlying document type.  If ``fields`` 
-            is not ``None``, only the fields in ``fields`` will be checked.
-        '''
-        try:
-            self.type.validate_unwrap(value, fields=fields, session=session)
-        except BadValueException, bve:
-            self._fail_validation(value, 'Bad value for DocumentField field', cause=bve)
-
 class BadIndexException(Exception):
     pass
 
@@ -688,37 +597,31 @@ class Index(object):
         collection.ensure_index(self.components, unique=self.__unique, 
             drop_dups=self.__drop_dups, **extras)
         return self
-        
-class Proxy(object):
-    def __init__(self, name):
-        self.name = name
-    def __get__(self, instance, owner):
-        if instance is None:
-            return getattr(owner, self.name)
-        session = instance._get_session()
-        ref = getattr(instance, self.name)
-        if ref is None:
-            return None
-        return session.dereference(ref)
-    def __set__(self, instance, value):
-        assert instance is not None
-        setattr(instance, self.name, value)
 
-class IProxy(object):
-    def __init__(self, name, ignore_missing):
-        self.name = name
-        self.ignore_missing = ignore_missing
-    def __get__(self, instance, owner):
-        if instance is None:
-            return getattr(owner, self.name)
-        session = instance._get_session()
-        def iterator():
-            for v in getattr(instance, self.name):
-                if v is None:
-                    yield v
-                    continue
-                value = session.dereference(v, allow_none=self.ignore_missing)
-                if value is None and self.ignore_missing:
-                    continue
-                yield value
-        return iterator()
+class Value(object):
+    def __init__(self, field, document, from_db=False, extra=False,
+                 retrieved=True):
+        # Stuff
+        self.field = field
+        self.doc = document
+        self.value = None
+        
+        # Flags
+        self.from_db = from_db
+        self.set = False
+        self.extra = extra
+        self.dirty = False
+        self.retrieved = retrieved
+        self.update_op = None
+    def clear_dirty(self):
+        self.dirty = False
+        self.update_op = None
+
+    def delete(self):
+        self.value = None
+        self.set = False
+        self.dirty = True
+        self.from_db = False
+        self.update_op = '$unset'
+
+
